# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

from pathlib import Path
import numpy as np

import xobjects as xo
import xpart as xp

from .version import SimVersion

class SimState(xo.HybridClass):
    _xofields = {
        'version':   SimVersion,    # This HAS to be the first field!
        'i_turn':    xo.Int64,
        'size':      xo.Int64,
        'particles': xp.Particles,
    }

    def __init__(self, **kwargs):
        if '_xobject' not in kwargs:
            kwargs['version'] = SimVersion()
        # TODO: can we set size at construction?
        super().__init__(**kwargs)

    @classmethod
    def from_binary(cls, filename, offset=0):
        # Read binary
        filename = Path(filename)
        with filename.open('rb') as fid:
            state_bytes = fid.read()
        buffer_data = xo.ContextCpu().new_buffer(capacity=len(state_bytes))
        buffer_data.buffer[:] = np.frombuffer(state_bytes, dtype=np.int8)
        # Cast to SimVersion to verify versions of xsuite packages
        version_offset = -1
        for field in cls._XoStruct._fields:
            if field.name == 'version':
                version_offset = field.offset
        if version_offset == -1:
            raise ValueError("No xofield 'version' found in SimState!")
        sim_ver_xobject = SimVersion._XoStruct._from_buffer(buffer=buffer_data, offset=offset+version_offset)
        sim_ver = SimVersion(_xobject=sim_ver_xobject)
        sim_ver.assert_version()
        # Retrieve simulation state
        sim_state_xobject = cls._XoStruct._from_buffer(buffer=buffer_data, offset=offset)
        sim_state_out = cls(_xobject=sim_state_xobject)
        return sim_state_out




# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import os, subprocess
from pathlib import Path
from time import sleep
import random

from .tools import log_debug, log_info, log_error


# Functions to work with the EOS file system
# ==========================================
EOS_MGM_URL = 'root://eosuser.cern.ch'
eos_env = {**os.environ, 'EOS_MGM_URL': EOS_MGM_URL}


def is_xrdcp_installed():
    try:
        # Run the xrdcp command with the "--version" flag to check if it's installed
        cmd = subprocess.run(["xrdcp", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
        if cmd.returncode == 0:
            return True
        else:
            return False
    except subprocess.CalledProcessError:
        # xrdcp is not installed or returned an error
        return False


def _eos_path(path):
    path = Path(path).expanduser().resolve()
    return path.as_posix().replace('home-','user/')

def _xrd_path(path):
    path = Path(path).expanduser().resolve()
    return EOS_MGM_URL + '/' + _eos_path(path)

def eos_glob(pattern, path, is_server=False):
    try:
        path = Path(path).expanduser().resolve()
        cmd = subprocess.run(['eos', 'find', '-name', pattern, _eos_path(path)],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=eos_env)
        if cmd.returncode != 0 or cmd.stdout == b'':
            return []
        files = [Path(ss).resolve() for ss in cmd.stdout.decode('UTF-8').strip().split('\n')]
        return [f for f in files if f.stat().st_size!=0]
    except Exception as e:
        log_error(f"Failed eos_glob {pattern} in {path}!\n", e, is_server=is_server)
        return []

def eos_exists(path, is_server=False):
    try:
        path = Path(path).expanduser().resolve()
        cmd = subprocess.run(['eos', 'find', '-name', path.name, _eos_path(path.parent)],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=eos_env)
        if cmd.returncode != 0 or cmd.stdout == b'':
            return False
        return Path(cmd.stdout.decode('UTF-8').strip()).resolve().stat().st_size!=0
    except Exception as e:
        log_error(f"Failed eos_exists for {path}!\n", e, is_server=is_server)
        return False

def eos_rm(path, is_server=False):
    try:
        path = Path(path).expanduser().resolve()
        cmd = subprocess.run(['eos', 'rm', _eos_path(path)], stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=eos_env)
        if cmd.returncode == 0:
            return 0
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            log_error(f"Failed eos_rm for {path}!\n{stderr}", is_server=is_server)
            return 1
    except Exception as e:
        log_error(f"Failed eos_rm for {path}!\n", e, is_server=is_server)
        return 1


# Always use this file by file, not for a list of files
def cp_from_eos(source, target, maximum_trials=10, wait=2.7, is_server=False):
    try:
        target = Path(target).expanduser().resolve()
        for i in range(maximum_trials):
            cmd = subprocess.run(['xrdcp', '--cksum', 'adler32', _eos_path(source), target.as_posix()], env=eos_env)
            if cmd.returncode == 0 and target.exists() and target.stat().st_size!=0:
                return 0
            if i != maximum_trials-1:
                log_debug(f"Failed to copy {str(source)} to {str(target)}!\nRetrying ({i})..", is_server=is_server)
                sleep(abs(wait+random.normalvariate(0, 0.1*wait)))
        log_error(f"Failed to copy {str(source)} to {str(target)}!", is_server=is_server)
        if not target.exists() or target.stat().st_size==0:
            log_error(f"Command succeeds but destination file is not created!", is_server=is_server)
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            log_error(f"Command failed: {stderr}", is_server=is_server)
        log_error("\nGiving up.", is_server=is_server)
        return 1
    except Exception as e:
        log_error(f"Failed to copy {str(source)} to {str(target)}!\n", e, is_server=is_server)
        return 1

def mv_from_eos(source, target, maximum_trials=10, wait=2.7, is_server=False):
    if not cp_from_eos(source, target, maximum_trials, wait, is_server=is_server):   # returncode 0 means success
        eos_rm(source, is_server=is_server)


# Always use this file by file, not for a list of files
def cp_to_eos(source, target, maximum_trials=10, wait=2.7, is_server=False):
    try:
        source = Path(source).expanduser().resolve()
        target = Path(target, source.name).expanduser().resolve()
#         target = Path(target).expanduser().resolve()
        for i in range(maximum_trials):
            cmd = subprocess.run(['xrdcp', '--cksum', 'adler32', source.as_posix(), _eos_path(target)], env=eos_env)
            if cmd.returncode == 0 and eos_exists(target):
                return 0
            if i != maximum_trials-1:
                log_debug(f"Failed to copy {str(source)} to {str(target)}!\nRetrying ({i})..", is_server=is_server)
                sleep(abs(wait+random.normalvariate(0, 0.1*wait)))
        log_error(f"Failed to copy {str(source)} to {str(target)}!", is_server=is_server)
        if not eos_exists(target):
            log_error(f"Command succeeds but destination file is not created!", is_server=is_server)
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            log_error(f"Command failed: {stderr}", is_server=is_server)
        log_error("\nGiving up.", is_server=is_server)
        return 1
    except Exception as e:
        log_error(f"Failed to copy {str(source)} to {str(target)}!\n", e, is_server=is_server)
        return 1

def mv_to_eos(source, target, maximum_trials=10, wait=2.7, is_server=False):
    if not cp_to_eos(source, target, maximum_trials, wait, is_server=is_server):   # returncode 0 means success
        source.unlink()



# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import json
from pathlib import Path

from .general import _pkg_root
from .user import update_user_data
from .server import server_account
from .server.paths import dropdir
from .server.eos import eos_exists, eos_rm, cp_to_eos
from .server.afs import afs_add_acl


user_data_file = _pkg_root / 'user_data.json'


def _create_json(user, folder):
    user_file = _pkg_root / f'register_{user}.json'
    if list(folder.parents)[-2].as_posix() == '/eos':
        domain = 'eos'
    elif list(folder.parents)[-3].as_posix() == '/afs/cern.ch':
        domain = 'afs'
    else:
        raise ValueError(f"Folder {folder} not on EOS nor AFS! "
                        + "The xboinc server will not be able to reach it.")
    data = {'user': user, 'folder': folder.as_posix(), 'domain': domain}
    with user_file.open('w') as f:
        json.dump(data, f)
    return user_file, data


def _set_rights(folder, domain):
    if domain == 'eos':
        raise NotImplementedError("Ask CERN IT for explanation...")
    else:
        afs_add_acl(server_account, folder)


def register(user, folder):
    folder = Path(folder).expanduser().resolve()
    if not folder.is_dir():
        raise ValueError(f"Folder {folder} not found or not a folder (or no permissions)!")
    user_file, data = _create_json(user, folder)
    try:
        _set_rights(folder, data['domain'])
    except Exception as e:
        user_file.unlink()
        raise Exception(e)
    if eos_exists(dropdir / user_file.name):
        eos_rm(dropdir / user_file.name)
        print("Replaced existing registration file on server dropdir.")
    if cp_to_eos(user_file, dropdir, maximum_trials=1):  # returncode 1 means error
        user_file.unlink()
        raise Exception(f"Failed to copy register file to server dropdir.\n"
                       + "Do you have permissions?\n"
                       + "If not, add yourself to the CERN xboinc-submitters egroup.")
    user_file.unlink()
    update_user_data(user, data)



# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

from pathlib import Path
import logging as lg
import traceback
import tarfile
import datetime



def timestamp(ms=False):
    ms = -3 if ms else -7
    return datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S.%f")[:ms]


def log_debug(message, is_server=False):
    if is_server:
        lg.debug(f"{timestamp(ms=True)}:  {message}")
    else:
        print(message)

def log_info(message, is_server=False):
    if is_server:
        lg.info(f"{timestamp(ms=True)}:  {message}")
    else:
        print(message)

def log_error(message, e=None, is_server=False):
    if is_server:
        stack = '' if e is None else f"\n{traceback.format_exc()}\n{e}"
        lg.error(f"{timestamp(ms=True)}:  {message}{stack}")
    else:
        raise Exception(message, e)

def untar(filename, is_server=False):
    try:
        filename = Path(filename)
        thisdir = filename.parent / filename.stem[:-4]
        thisdir.mkdir(parents=True, exist_ok=True)
        with tarfile.open(filename, 'r:gz') as fid:
            fid.extractall(path=thisdir)
        filename.unlink()
        log_debug(f"Extracted {filename}.", is_server)

    except Exception as e:
        log_error(f"Failed extracting {filename}", e, is_server)



# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import json
from pathlib import Path

from .general import _pkg_root


user_data_file = _pkg_root / 'user_data.json'
if not user_data_file.exists():
    with user_data_file.open('w') as fid:
        json.dump({}, fid)


def update_user_data(user, data):
    with user_data_file.open('r') as fid:
        userdict = json.load(fid)
    data.pop('user', None)
    if user in userdict:
        userdict[user].update(data)
    else:
        userdict[user] = data
    with user_data_file.open('w') as fid:
        json.dump(userdict, fid)

def get_user_data(user):
    with user_data_file.open('r') as fid:
        userdict = json.load(fid)
    return userdict[user]

def get_folder(user):
    return Path(get_user_data(user)['folder']).resolve()

def get_domain(user):
    return get_user_data(user)['domain']



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['xboinc', 'xboinc.executable', 'xboinc.server', 'xboinc.simulation_io']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.0',
 'xcoll==0.2.5',
 'xdeps==0.4.2',
 'xfields==0.13.1',
 'xobjects==0.2.8',
 'xpart==0.15.3',
 'xtrack==0.42.0']

setup_kwargs = {
    'name': 'xboinc',
    'version': '0.1.1',
    'description': 'Xsuite BOINC interface',
    'long_description': '# XBOINC\n\n## Python package for submitting track simulations in particle accelerators with BOINC system\n\n## INSTRUCTIONS FOR WINDOWS:\n\n### How to install boinc client/manager on Windows\n\nUse the installer at: https://boinc.berkeley.edu/download.php\n\n\n### Compile boinc api and examples from source on Windows\n\nWe use MSys2, in particular **MINGW64**\n\nFollow the instruction at https://www.msys2.org to install MSYS2 and GCC:\n```\npacman -S mingw-w64-x86_64-gcc\npacman -Suy\npacman -Suy  # you need indeed twice!\n```\n\nInstall git and python packages:\n```\npacman -S git\npacman -Sy mingw-w64-x86_64-python-numpy\npacman -Sy mingw-w64-x86_64-python-scipy\npacman -Sy mingw-w64-x86_64-python-pandas\npacman -Sy mingw-w64-x86_64-python-pip\npacman -Sy mingw-w64-x86_64-python-wheel\npacman -Sy vim\n```\n\n#### Compile BOINC\n```\npacman -S autotools\ncd\ngit clone https://github.com/BOINC/boinc\ncd boinc\nexport BOINC_DIR=$PWD\n \n./_autosetup\n \n./configure --disable-server --disable-client --disable-manager --disable-gui\n# This one takes very long\n \nmake\n# This also takes quite long\n```\n\n### Generate xboinc source and input files on windows\n\n```\ncd \nmkdir xsuite_packages\ncd xsuite_packages\ngit clone https://github.com/xsuite/xobjects.git\ngit clone https://github.com/xsuite/xpart.git\ngit clone https://github.com/xsuite/xtrack.git\ngit clone https://github.com/xsuite/xfields.git\ngit clone https://github.com/xsuite/xboinc.git\n\npip install --no-deps --no-build-isolation -e xtrack\npip install --no-deps --no-build-isolation -e xobjects\npip install --no-deps --no-build-isolation -e xpart\npip install --no-deps --no-build-isolation -e xfields\npip install --no-deps --no-build-isolation -e xboinc\n```\n\n#### Compile and generate input\n\n\n```bash\ncd xboinc/examples/000_build_executable\npython 000_build_executable.py\nbash 001msys2_compile_executable.sh\ncd ../002_lhc\npython 001_build_input.py\ncd ../003_boinc\necho $BOINC_DIR # Check if points into the BOINC source directory \npython 000_build_executable.py\nmake\n```\n\n#### Run/test the executable\n```\ncp ../002_lhc/xboinc_input.bin .\nchmod +x xboinc_executable\n./xboinc_executable\n```\n\n#### Running the test application with boinc server\nWe still work in MSYS2 (MINGW64)\n\nIn order to have it working we needed to have the file structure in C:\\PrgramData\\Boinc\n\n```\ncd /c/ProgramData/BOINC # This in WINDOWS is hidden\ncp ~/xsuite_packages/xboinc/examples/003_boinc/client_state_save.xml .\n```\nthen\n\n```\nnano account_test.xml \n```\n\nwrite\n```\n<account>\n    <master_url>http://test.test</master_url>\n    <project_name>test_project</project_name>\n</account>\n```\nthen \n```nano cc_config.xml```\nwrite\n```\n<cc_config>\n    <options>\n        <skip_cpu_benchmarks/>\n        <unsigned_apps_ok/>\n    </options>\n</cc_config>\n```\nThen make folder structure\n```\nmkdir projects/test.test\ncp ~/xsuite_packages/xboinc/examples/002_lhc/xboinc_input.bin projects/test.test/input.bin\ncp ~/xsuite_packages/xboinc/examples/003_boinc/xboinc_executable.exe projects/test.test/xboinc_executable\n```\n\nIn one terminal:\n`cp client_state_save.xml client_state.xml; ./../../Program\\ Files/BOINC/boinc.exe`\n\nIn another:\n`/c/Program\\ Files/BOINC/boincmgr.exe`\n\nIn case of problems stop boing client in the status bar:\n\n<img width="427" alt="image" src="https://user-images.githubusercontent.com/8049893/191972091-13fee31a-9dc8-459e-9f3f-3224c09bec47.png">\n\nYou can suspend using the activity menu:\n\n<img width="785" alt="image" src="https://user-images.githubusercontent.com/8049893/191973210-64e0c156-9565-41d9-9e8c-8e5eeef19b2b.png">\n\nYou can see the checkpoints in:\n\n<img width="835" alt="image" src="https://user-images.githubusercontent.com/8049893/191973307-498ce522-7542-4975-9e18-702bf449abd3.png">\n\nYou can restart from the same menu. \n\n\nWhen the job is finished, the result are saved in "projects/test.test/output.bin"\n\n\n## INSTRUCTIONS FOR UBUNTU:\n\n### How to install boinc client/manager on Ubuntu\nhttps://boinc.berkeley.edu/wiki/Installing_BOINC_on_Ubuntu\n\nBasically only:\n```\nsudo apt-get install aptitude\nsudo aptitude install boinc-client boinc-manager\n```\n#### Compile boinc api and examples from source on Ubuntu\n\nOn a fresh ubuntu installation I neede to:\n\n```\nsudo apt install make\nsudo apt install m4\nsudo apt install pkg-config\nsudo apt install autoconf\nsudo apt install libtool\nsudo apt install g++\n```\n\n```\ncd\ngit clone https://github.com/BOINC/boinc\ncd boinc\nexport BOINC_DIR=$PWD\n./_autosetup\n```\nFrom \nhttps://boinc.berkeley.edu/trac/wiki/BuildSystem\n\n"If you\'re developing or porting a BOINC application, you need only the API: "\n```\n./configure --disable-server --disable-client --disable-manager --disable-gui\n# This one takes very long\n```\nand then\n```\nmake\n# This also takes quite long\n```\n\n### Installing xsuite packages\n```\ncd\nwget https://repo.anaconda.com/miniconda/Miniconda3-latest-Linux-x86_64.sh\nbash Miniconda3-latest-Linux-x86_64.sh\nsource miniconda3/bin/activate\npip install numpy scipy matplotlib pandas ipython pytest\nmkdir xsuite_packages\ncd xsuite_packages\n\ngit clone https://github.com/xsuite/xobjects\ngit clone https://github.com/xsuite/xdeps\ngit clone https://github.com/xsuite/xpart\ngit clone https://github.com/xsuite/xtrack\ngit clone https://github.com/xsuite/xfields\ngit clone https://github.com/xsuite/xboinc\n\npip install -e xobjects\npip install -e xdeps\npip install -e xpart\npip install -e xtrack\npip install -e xfields\npip install -e xboinc\n```\n\n#### Compile and generate input\n\n```bash\ncd xboinc/examples/000_build_executable\npython 000_build_executable.py\nbash 001msys2_compile_executable.sh\ncd ../002_lhc\npython 001_build_input.py\ncd ../003_boinc\necho $BOINC_DIR # Check if points into the BOINC source directory \npython 000_build_executable.py\nmake\n```\n\n#### Run/test the executable\n```\ncp ../002_lhc/xboinc_input.bin .\nchmod +x xboinc_executable\n./xboinc_executable\n```\n\n### Running the test application with boinc server\n```\ncd\nmkdir xboinc_test\ncp ~/xsuite_packages/xboinc/examples/003_boinc/client_state_save.xml .\n```\nthen\n\n```\nvim account_test.xml \n```\n\nwrite\n```\n<account>\n    <master_url>http://test.test</master_url>\n    <project_name>test_project</project_name>\n</account>\n```\nthen \n```vim cc_config.xml```\nwrite\n```\n<cc_config>\n    <options>\n        <skip_cpu_benchmarks/>\n        <unsigned_apps_ok/>\n    </options>\n</cc_config>\n```\n\nAdd the same GUI RPC passwords in\n```\nvim gui_rpc_auth.cfg\n```\nand in\n```\nvim /etc/boinc-client/gui_rpc_auth.cfg\n```\n\nThen make folder structure\n```\nmkdir projects/test.test\ncp ~/xsuite_packages/xboinc/examples/002_lhc/xboinc_input.bin projects/test.test/input.bin\ncp ~/xsuite_packages/xboinc/examples/003_boinc/xboinc_executable.exe projects/test.test/xboinc_executable\n```\n\nIn one terminal:\n`cp client_state_save.xml client_state.xml; boinc --gui_rpc_port 31417`\n\nIn another:\n`boincmgr --gui_rpc_port 31417`\n\n!! Before running BOINC client or manager, kill the previous BOINC session\n```\nsudo /etc/init.d/boinc-client stop\n```\n',
    'author': 'Frederik F. Van der Veken',
    'author_email': 'frederik@cern.ch',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/xsuite/xboinc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import subprocess
from pathlib import Path

from .tools import log_debug, log_info, log_error


def afs_add_acl(user, folder, acl='rlwik', is_server=False):
    try:
        folder = Path(folder).resolve()
        if folder.is_dir():
            cmd = subprocess.run(['fs', 'sa', folder, user, acl],
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            if cmd.returncode == 0:
                log_info(f"Set ACL '{acl}' on folder {folder} for user {user}.", is_server=is_server)
                return 0
            else:
                stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
                log_error(f"Failed to set ACL '{acl}' on {folder} for user {user}.\n{stderr}", is_server=is_server)
                return 1
        else:
            log_error(f"Folder {folder} not found. Do you have access permissions?", is_server=is_server)
            return 1
    except Exception as e:
        log_error(f"Failed to set ACL '{acl}' on {folder} for user {user}.\n{stderr}", e, is_server=is_server)
        return 1


def afs_remove_acl(user, folder, is_server=False):
    try:
        folder = Path(folder).resolve()
        if folder.is_dir():
            cmd = subprocess.run(['fs', 'sa', folder, user, 'none'],
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            if cmd.returncode == 0:
                log_info(f"Removed ACL on folder {folder} for user {user}.", is_server=is_server)
                return 0
            else:
                stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
                log_error(f"Failed to remove ACL on {folder} for user {user}.\n{stderr}", is_server=is_server)
                return 1
        else:
            log_error(f"Folder {folder} not found. Do you have access permissions?", is_server=is_server)
            return 1
    except Exception as e:
        log_error(f"Failed to remove ACL on {folder} for user {user}.\n{stderr}", e, is_server=is_server)
        return 1


def afs_print_acl(folder, is_server=False):
    try:
        folder = Path(folder).resolve()
        if folder.is_dir():
            cmd = subprocess.run(['fs', 'la', folder],
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            if cmd.returncode == 0:
                stdout = cmd.stdout.decode('UTF-8').strip().split('\n')
                log_info(f"{stdout}", is_server=is_server)
                return 0
            else:
                stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
                log_error(f"Failed to remove ACL on {folder} for user {user}.\n{stderr}", is_server=is_server)
                return 1
        else:
            log_error(f"Folder {folder} not found. Do you have access permissions?", is_server=is_server)
            return 1
    except Exception as e:
        log_error(f"Failed to remove ACL on {folder} for user {user}.\n{stderr}", e, is_server=is_server)
        return 1


def afs_rm(path, is_server=False):
    try:
        path = Path(path).expanduser().resolve()
        cmd = subprocess.run(['rm', path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if cmd.returncode == 0:
            return 0
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            log_error(f"Failed afs_rm for {path}!\n{stderr}", is_server=is_server)
            return 1
    except Exception as e:
        log_error(f"Failed afs_rm for {path}!\n", e, is_server=is_server)
        return 1


# Always use this file by file, not for a list of files
def cp_from_afs(source, target, maximum_trials=10, wait=2.7, is_server=False):
    try:
        source = Path(source).expanduser().resolve()
        target = Path(target).expanduser().resolve()
        for i in range(maximum_trials):
            cmd = subprocess.run(['cp', source.as_posix(), target.as_posix()])
            if cmd.returncode == 0 and target.exists() and target.stat().st_size!=0:
                return 0
            if i != maximum_trials-1:
                log_debug(f"Failed to copy {str(source)} to {str(target)}!\nRetrying ({i})..", is_server=is_server)
                sleep(abs(wait+random.normalvariate(0, 0.1*wait)))
        log_error(f"Failed to copy {str(source)} to {str(target)}!", is_server=is_server)
        if not target.exists() or target.stat().st_size==0:
            log_error(f"Command succeeds but destination file is not created!", is_server=is_server)
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            log_error(f"Command failed: {stderr}", is_server=is_server)
        log_error("\nGiving up.", is_server=is_server)
        return 1
    except Exception as e:
        log_error(f"Failed to copy {str(source)} to {str(target)}!\n", e, is_server=is_server)
        return 1

def mv_from_afs(source, target, maximum_trials=10, wait=2.7, is_server=False):
    if not cp_from_afs(source, target, maximum_trials, wait, is_server=is_server):   # returncode 0 means success
        afs_rm(source, is_server=is_server)


# Always use this file by file, not for a list of files
def cp_to_afs(source, target, maximum_trials=10, wait=2.7, is_server=False):
    try:
        source = Path(source).expanduser().resolve()
        target = Path(target).expanduser().resolve()
        for i in range(maximum_trials):
            cmd = subprocess.run(['cp', source.as_posix(), target.as_posix()])
            if cmd.returncode == 0 and target.exists() and target.stat().st_size!=0:
                return 0
            if i != maximum_trials-1:
                log_debug(f"Failed to copy {str(source)} to {str(target)}!\nRetrying ({i})..", is_server=is_server)
                sleep(abs(wait+random.normalvariate(0, 0.1*wait)))
        log_error(f"Failed to copy {str(source)} to {str(target)}!", is_server=is_server)
        if not target.exists() or target.stat().st_size==0:
            log_error(f"Command succeeds but destination file is not created!", is_server=is_server)
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            log_error(f"Command failed: {stderr}", is_server=is_server)
        log_error("\nGiving up.", is_server=is_server)
        return 1
    except Exception as e:
        log_error(f"Failed to copy {str(source)} to {str(target)}!\n", e, is_server=is_server)
        return 1

def mv_to_afs(source, target, maximum_trials=10, wait=2.7, is_server=False):
    if not cp_to_afs(source, target, maximum_trials, wait, is_server=is_server):   # returncode 0 means success
        source.unlink()
